#include <stdlib.h>
#include <list>
#include <map>
#include <iostream>
#include <time.h>
#include <omp.h>
#include "CGP.h"
#include "RecycleBin.h"
#include "CGPGraphViz.h"
using namespace CartesianGeneticProgramming;
using namespace std;

#define MODEL_VERSION ("2011-12-18a")

/*
Versions:
Wishlist
	- TODO Accept an iteration count on the command line
	- TODO Randomly pace the command packets in the fitness test data schedule.
	- TODO Video and telemetry data in the fitness test data schedule should have regular but not identical timings, making periods of rapid sending and idle periods.

2011-12-18a
	- Experiment with transceiver population of 4 without an attacker.  Does it produce results as good as the population of 8?
	- Set the attacker population size independent of the transceiver population.
	- Experiement with transceiver population of 4 and attacker population of 4, for 16 required worlds.
	- All attacker variants must play all transceiver variants.
	- Give small reward for the attacker when it sends packets. This is to combat a mute attacker.

2011-12-17a
	- Reduce the instruction set by removing the bitwise functions.
	- Allow only integer constants in functions by defining CGP_INTEGER_CONSTANTS
	- Fix fitness for non-chromosome case so fitness with and without chromosomes can be compared.
	- Attacker chromosome and model.  The attacker is turned on by defining USE_ATTACKER.  Previous code worked as if USE_ATTACKER was not defined.
	- Attacker scoring.
	- Attacker best generation and individual.
	- Log USE_ATTACKER.
	- Log attacker fitness and generation information.

2011-12-15b
	- Added choice between evolving individuals or chromosomes.  USE_CHROMOSOMES controls this setting.
		Prior code behaved as if USE_CHROMOSOMES were defined, using chromosonal evolution.

2011-12-15a
	- Added FITNESS_ITERATION_COUNT, default to 10 for initial testing. Code behaved as if set to 1 previously.
		The goal of this is to average out the randomness of the fitness tests, which must have randomness to test generalization.
	- Randomized the starting sequence number for packets, from 1 to 5000.
	- Randomized the sequence number increment between packets, from 1 to 10.
	- Removed penalty for a lost packet when Space loses a video or telemetry packet.
	- Automatically discard data bearing packets with inappropriate data types. Rovers ignore video and telemetry. Earth ignores commands.
		These automatically discarded packets are never presented to the CGP and don't affect scoring.
	- Space corrupts packets and some signature types can recover from this corruption.  This affects validation as well.
		Without this feature, packet signature type is meaningless.
		The goal is to choose a signature type that fits the data type and to device a signature selection technique that will confuse an attacker.

2011-12-13a
	- Lowered the mutation rate for testing.
	- Accelerate mutation rate if no improvement has occured in MUTATION_ACCELERATION_GENERATIONS generations.
		MUTATION_RATE and HIGH_MUTATION_RATE are used for the base and accelerated rates respectively.
*/

template <class T>
void DeleteAll(list<T *> &all) {
	while(!all.empty()) {
		T *t = all.front();
		all.pop_front();
		delete t;
	}
}

#ifdef _DEBUG
#define FITNESS_ITERATION_COUNT (1)
#define POPULATION_SIZE (2)
#define DATA_SCHEDULE_COUNT (5)
#define ATTACKER_POPULATION_SIZE (2)
#else
#define FITNESS_ITERATION_COUNT (10)
//#define POPULATION_SIZE (8)
#define POPULATION_SIZE (4)
#define DATA_SCHEDULE_COUNT (100)

#define ATTACKER_POPULATION_SIZE (4)
#endif


//#define USE_CHROMOSOMES

#define USE_ATTACKER

#define ATTACKER_STATE_COUNT (4)

#define MUTATION_ACCELERATION_GENERATIONS (10000)
//#define MUTATION_RATE (0.08f)
//#define HIGH_MUTATION_RATE (0.15f)
#define MUTATION_RATE (0.04f)
#define HIGH_MUTATION_RATE (0.04f)

#define CONTROL_VALUE_COUNT (2)
#define TRANSMIT_QUEUE_COUNT (5)
#define HISTORY_QUEUE_COUNT (10)
#define INTERNAL_STATE_COUNT (2)
#define TRANSMIT_HISTORY_TIMEOUT_SECONDS (20.0f)
#define TRANSMIT_SECONDS (0.05f)
#define SPACE_SECONDS (2.5f)
#define SPACE_LOSS_FRACTION (0.05)
#define SPACE_REPLAY_FRACTION (0.05)
#define SPACE_ECHO_FRACTION (0.05)
#define SPACE_CORRUPT_FRACTION (0.05)
#define SPACE_UNREPAIRABLE_CORRUPT_FRACTION (0.125)
#define RUN_TIME_AFTER_LAST_DATA_SECONDS (TRANSMIT_HISTORY_TIMEOUT_SECONDS*5)

#define ATTACKER_IDLE_SECONDS (0.25f)

#define VALIDATION_DATA_SCHEDULE_COUNT (200)
#define VALIDATION_SPACE_LOSS_FRACTION (0.125)
#define VALIDATION_SPACE_REPLAY_FRACTION (0.125)
#define VALIDATION_SPACE_ECHO_FRACTION (0.125)
#define VALIDATION_SPACE_CORRUPT_FRACTION (0.125)
#define VALIDATION_SPACE_UNREPAIRABLE_CORRUPT_FRACTION (0.25)

// Optimize for size by including the active gene count into the fitness
// 0 == Never
// 1 == Always
// 2 == Only if the score is otherwise perfect
#define OPTIMIZE_FOR_SIZE (2)

#define PENALTY_ACCEPTED_BAD_SIGNATURE (500)
#define PENALTY_ACCEPTED_ATTACKER_PACKET (1000)
#define PENALTY_ACCEPTED_UNKNOWN_DATA_LUMP (500)
#define PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE (500)
#define PENALTY_NEVER_ACCEPTED_DATA_LUMP (1000)
#define PENALTY_ACCEPTED_DATA_LUMP_WITH_WRONG_DATA_TYPE (1000)
#define PENALTY_COMMAND_PACKET_GOT_THROUGH (500)

//#define PENALTY_ATTACKER_SENT_TO_EARTH (100)
//#define PENALTY_ATTACKER_SENT_TO_MOON (250)

#define PENALTY_ATTACKER_SENT_TO_EARTH (-1)
#define PENALTY_ATTACKER_SENT_TO_MOON (-1)


#define INFINITE_CLOCK (1e+36f)

enum DataSource {
	DS_Unknown=0,
	DS_Earth,
	DS_Rover,
	DS_Attacker
};

enum DataType {
	DT_None = 0,
	DT_Video = 1,
	DT_Telemetry = 2,
	DT_Command = 3
};

class DataLump {
public:
	float id;
	DataType dataType;
	DataSource whichTransceiver;
	float creationTime;
	bool hasBeenDequeued;
	int acceptanceCounter;
	bool wasDiscardedBySpace;

public:
	DataLump()
		: id(0), dataType(DT_None), whichTransceiver(DS_Unknown), creationTime(0), hasBeenDequeued(false), acceptanceCounter(0), wasDiscardedBySpace(false) {
	}
};

typedef struct {
	DataType dataType;
	float dataID;
	DataSource source;
	float signatureType;
	bool signatureOk;
	float controlValues[CONTROL_VALUE_COUNT];
	float sequenceNumber; // 0 for control-only packets
	float whenExpiresInHistory;
} Packet;

class TimedPacket {
public:
	Packet *packet;
	float whenArrives;

public:
	TimedPacket()
		: packet(NULL), whenArrives(0) {
	}

	TimedPacket(Packet *packetToUse, float whenArrivesToUse)
		: packet(packetToUse), whenArrives(whenArrivesToUse) {
	}

	~TimedPacket() {
		if(packet) {
			delete packet;
			packet = NULL;
		}
	}
};

class Attacker {
public:
#if ATTACKER_STATE_COUNT > 0
	float attackerState[ATTACKER_STATE_COUNT];
#endif
	list<Packet *> fromSpace;
	list<Packet *> fromEarth;
	//list<TimedPacket *> transmitQueue;

	map<int,float> signatureTypeCrackCounter;
	Packet *lastSentPacket;
	float nextStepTime;

public:
	Attacker()
		: lastSentPacket(NULL), nextStepTime(0) {
	}

	~Attacker() {
		DeleteAll(fromSpace);
		DeleteAll(fromEarth);
		//DeleteAll(transmitQueue);
		if(lastSentPacket) {
			delete lastSentPacket;
			lastSentPacket = NULL;
		}
	}

	void OnSignatureTypeSeen(int signatureType, DataType dataType, RandomNumberSource &randomSource) {
		if(signatureType > 0) {
			if(signatureTypeCrackCounter.find(signatureType) == signatureTypeCrackCounter.end()) {
				signatureTypeCrackCounter[signatureType] = (float)(randomSource.NextInt(5000) + 1);
			}
			float counter = signatureTypeCrackCounter[signatureType];
			if(counter > 0) {
				switch(dataType) {
					case DT_Video:
						counter -= 50;
						break;
					case DT_Telemetry:
						counter -= 10;
						break;
					case DT_Command:
						counter -= 1;
						break;
					default:
						counter -= 1;
				}
				signatureTypeCrackCounter[signatureType] = counter;
			}
		}
	}

	bool IsSignatureTypeCracked(int signatureType) const {
		if(signatureType <= 0) {
			return true;
		}
		map<int,float>::const_iterator it = signatureTypeCrackCounter.find(signatureType);
		if(it != signatureTypeCrackCounter.end()) {
			return (*it).second <= 0;
		}
		return false;
	}
};

class Transceiver {
public:
	float packetReceivedState;
#if INTERNAL_STATE_COUNT > 0
	float sharedState[INTERNAL_STATE_COUNT];
#else
	float sharedState[1];
#endif
	float nextSequenceNumber, sequenceIncrement, transmitterState;
	Packet *packetInTransmitter;
	float whenTransmitFinishes;
	float highestSequenceSeen;

	list<Packet *> transmitQueue; // front is oldest, new entries are pushed to the back
	list<Packet *> historyQueue; // front is oldest, new entries are pushed to the back

	DataSource location;

public:
	Transceiver(DataSource locationToUse)
			: location(locationToUse) {
		Initialize();
	}

	~Transceiver() {
		if(packetInTransmitter) {
			delete packetInTransmitter;
			packetInTransmitter = NULL;
		}
		DeleteAll(transmitQueue);
		DeleteAll(historyQueue);
	}

	void Initialize() {
		packetReceivedState = 0;
		transmitterState = 0;
		nextSequenceNumber = 1;
		sequenceIncrement = 1;
		for(int i=0;i<sizeof(sharedState)/sizeof(sharedState[0]);i++) {
			sharedState[i] = 0;
		}
		packetInTransmitter = NULL;
		whenTransmitFinishes = 0;
		highestSequenceSeen = nextSequenceNumber - sequenceIncrement;
	}

	void Randomize(RandomNumberSource &randomSource) {
		nextSequenceNumber = (float)(1 + randomSource.NextInt(5000));
		sequenceIncrement = (float)(1 + randomSource.NextInt(10));
		highestSequenceSeen = nextSequenceNumber - sequenceIncrement;
	}

	float GetOldestSequenceInHistory() const {
		if(historyQueue.empty()) {
			return 0;
		}
		return historyQueue.front()->sequenceNumber;
	}
};

class SpacePacket {
public:
	Packet *packet;
	float whenArrives;

public:
	SpacePacket()
		: packet(NULL), whenArrives(0) {
	}

	SpacePacket(Packet *packetToUse, float whenArrivesToUse)
		: packet(packetToUse), whenArrives(whenArrivesToUse) {
	}

	~SpacePacket() {
		if(packet) {
			delete packet;
			packet = NULL;
		}
	}
};

class World1 {
public:
	RandomNumberSource randomSource;
	Transceiver earthHardware;
	Transceiver roverHardware;
	Attacker attacker;
	list<SpacePacket *> toRoverInSpaceQueue;
	list<SpacePacket *> toEarthInSpaceQueue;
	float now, stopTime;
	float attackerPacketReceivedFitness, packetReceivedFitness, dataReadyFitness, packetExpiredFitness;
	CGPIndividual *packetReceivedProgram;
	CGPIndividual *dataReadyProgram;
	CGPIndividual *packetExpiredProgram;
	CGPIndividual *attackerPacketReceivedProgram;

	RecycleBin<Packet,1000> packetBin;
	RecycleBin<SpacePacket,100> spacePacketBin;
	RecycleBin<TimedPacket,100> timedPacketBin;

	DataLump *dataSchedule;
	int dataScheduleCount;
	int nextDataIndex;

	CGPExecutor packetReceivedExecutor;
	CGPExecutor dataReadyExecutor;
	CGPExecutor packetExpiredExecutor;
	CGPExecutor attackerPacketReceivedExecutor;

	int spaceState;
	bool isValidation;

	// Statistics
	int numberOfDataLumpsAcceptedMoreThanOnce;
	int numberOfExtraAcceptances;
	int numberOfDataLumpsNeverAccepted;
	int numberOfCommandPacketsNeverAccepted;
	int numberOfVideoPacketsNeverAccepted;
	int numberOfTelemetryPacketsNeverAccepted;

	int numberOfCommandPacketsSent;
	int numberOfVideoPacketsSent;
	int numberOfTelemetryPacketsSent;

	int numberOfPacketsLostInSpace;

	int numberOfBadSignaturesAccepted;
	int numberOfAttackerPacketsAccepted;
	int numberOfUnknownDataAccepted;

	int numberOfPacketsSentByAttacker;
	int numberOfPacketsSentByRover;
	int numberOfPacketsSentByEarth;

public:
	World1()
		: earthHardware(DS_Earth), roverHardware(DS_Rover), now(0), stopTime(0), attackerPacketReceivedFitness(0),
			packetReceivedFitness(0), dataReadyFitness(0), packetExpiredFitness(0),
			packetReceivedProgram(NULL), dataReadyProgram(NULL), packetExpiredProgram(NULL),
			attackerPacketReceivedProgram(NULL),
			dataSchedule(NULL), dataScheduleCount(0), nextDataIndex(0), spaceState(0), isValidation(false),
			numberOfDataLumpsAcceptedMoreThanOnce(0),
			numberOfExtraAcceptances(0),
			numberOfDataLumpsNeverAccepted(0),
			numberOfCommandPacketsNeverAccepted(0),
			numberOfVideoPacketsNeverAccepted(0),
			numberOfTelemetryPacketsNeverAccepted(0),
			numberOfCommandPacketsSent(0),
			numberOfVideoPacketsSent(0),
			numberOfTelemetryPacketsSent(0),
			numberOfPacketsLostInSpace(0),
			numberOfBadSignaturesAccepted(0),
			numberOfAttackerPacketsAccepted(0),
			numberOfUnknownDataAccepted(0), 
			numberOfPacketsSentByAttacker(0), numberOfPacketsSentByRover(0), numberOfPacketsSentByEarth(0) {
		earthHardware.Randomize(randomSource);
		roverHardware.Randomize(randomSource);
	}

	~World1() {
		DeleteAll(toRoverInSpaceQueue);
		DeleteAll(toEarthInSpaceQueue);
		if(dataSchedule) {
			delete [] dataSchedule;
			dataSchedule = NULL;
		}
	}

	void CreateDataSchedule() {
		if(isValidation) {
			dataScheduleCount = VALIDATION_DATA_SCHEDULE_COUNT;
			dataSchedule = new DataLump[dataScheduleCount];
			nextDataIndex = 0;
			float t = 0;
			for(int i=0;i<dataScheduleCount;i++) {
				t += (float)(randomSource.Next()*3 + 0.001);
				dataSchedule[i].id = (float)(1 + i);
				dataSchedule[i].creationTime = t;
				dataSchedule[i].dataType = (DataType)(randomSource.NextInt(3) + 1);
				dataSchedule[i].whichTransceiver = dataSchedule[i].dataType == DT_Command? DS_Earth : DS_Rover;
			}
			stopTime = t + RUN_TIME_AFTER_LAST_DATA_SECONDS;
		} else {
			dataScheduleCount = DATA_SCHEDULE_COUNT;
			dataSchedule = new DataLump[dataScheduleCount];
			nextDataIndex = 0;
			float t = 0;
			for(int i=0;i<dataScheduleCount;i++) {
				t += (float)((((i+1)%5)*(0.15+((i+1)%11)*0.05))+0.01);
				dataSchedule[i].id = (float)(1 + i);
				dataSchedule[i].creationTime = t;
				dataSchedule[i].dataType = (DataType)((i % 3) + 1);
				dataSchedule[i].whichTransceiver = dataSchedule[i].dataType == DT_Command? DS_Earth : DS_Rover;
			}
			stopTime = t + RUN_TIME_AFTER_LAST_DATA_SECONDS;
		}
	}

	void CalculateFitnessScores(bool printFitnessDetails) {
		for(int i=0;i<dataScheduleCount;i++) {
			DataLump *d = &dataSchedule[i];
			if(d->hasBeenDequeued) {
				switch(d->dataType) {
					case DT_Command:
						numberOfCommandPacketsSent++;
						break;
					case DT_Video:
						numberOfVideoPacketsSent++;
						break;
					case DT_Telemetry:
						numberOfTelemetryPacketsSent++;
						break;
				}
			}
			if(d->acceptanceCounter > 1) {
				numberOfDataLumpsAcceptedMoreThanOnce++;
				numberOfExtraAcceptances += d->acceptanceCounter - 1;
			} else if(d->acceptanceCounter <= 0) {
				numberOfDataLumpsNeverAccepted++;
				switch(d->dataType) {
					case DT_Command:
						numberOfCommandPacketsNeverAccepted++;
						break;
					case DT_Video:
						if(d->wasDiscardedBySpace) {
							// These could never be resent, so don't penalize the CGP for something Space lost at random.
							numberOfDataLumpsNeverAccepted--;
						}
						numberOfVideoPacketsNeverAccepted++;
						break;
					case DT_Telemetry:
						if(d->wasDiscardedBySpace) {
							// These could never be resent, so don't penalize the CGP for something Space lost at random.
							numberOfDataLumpsNeverAccepted--;
						}
						numberOfTelemetryPacketsNeverAccepted++;
						break;
				}
			}
		}
		packetReceivedFitness += PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE * numberOfExtraAcceptances;
		dataReadyFitness += PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE * numberOfExtraAcceptances;
		packetExpiredFitness += PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE * numberOfExtraAcceptances;

		packetReceivedFitness += PENALTY_NEVER_ACCEPTED_DATA_LUMP * numberOfDataLumpsNeverAccepted;
		dataReadyFitness += PENALTY_NEVER_ACCEPTED_DATA_LUMP * numberOfDataLumpsNeverAccepted;
		packetExpiredFitness += PENALTY_NEVER_ACCEPTED_DATA_LUMP * numberOfDataLumpsNeverAccepted;

#if OPTIMIZE_FOR_SIZE == 2
		if(packetReceivedFitness <= 0) {
			packetReceivedFitness += packetReceivedProgram->activeGeneCount;
		}
		if(dataReadyFitness <= 0) {
			dataReadyFitness += dataReadyProgram->activeGeneCount;
		}
		if(packetExpiredFitness <= 0) {
			packetExpiredFitness += packetExpiredProgram->activeGeneCount;
		}
#elif OPTIMIZE_FOR_SIZE == 1
		packetReceivedFitness += packetReceivedProgram->activeGeneCount;
		dataReadyFitness += dataReadyProgram->activeGeneCount;
		packetExpiredFitness += packetExpiredProgram->activeGeneCount;
#elif OPTIMIZE_FOR_SIZE == 0
		// Never optimize for size
#else
#error
#endif

#ifndef USE_CHROMOSOMES
		float totalFitness = (packetReceivedFitness + dataReadyFitness + packetExpiredFitness) / 3;
		packetReceivedFitness = totalFitness;
		dataReadyFitness= totalFitness;
		packetExpiredFitness = totalFitness;
#endif

		if(printFitnessDetails) {
			std::cout << "t.startTrial" << std::endl;
			std::cout << "t.numberOfDataLumpsAcceptedMoreThanOnce=" << numberOfDataLumpsAcceptedMoreThanOnce << std::endl;
			std::cout << "t.numberOfExtraAcceptances=" << numberOfExtraAcceptances << std::endl;
			std::cout << "t.numberOfDataLumpsNeverAccepted=" << numberOfDataLumpsNeverAccepted << std::endl;
			std::cout << "t.numberOfCommandPacketsNeverAccepted=" << numberOfCommandPacketsNeverAccepted << std::endl;
			std::cout << "t.numberOfVideoPacketsNeverAccepted=" << numberOfVideoPacketsNeverAccepted << std::endl;
			std::cout << "t.numberOfTelemetryPacketsNeverAccepted=" << numberOfTelemetryPacketsNeverAccepted << std::endl;
			std::cout << "t.numberOfCommandPacketsSent=" << numberOfCommandPacketsSent << std::endl;
			std::cout << "t.numberOfVideoPacketsSent=" << numberOfVideoPacketsSent << std::endl;
			std::cout << "t.numberOfTelemetryPacketsSent=" << numberOfTelemetryPacketsSent << std::endl;
			std::cout << "t.numberOfPacketsLostInSpace=" << numberOfPacketsLostInSpace << std::endl;
			std::cout << "t.numberOfBadSignaturesAccepted=" << numberOfBadSignaturesAccepted << std::endl;
			std::cout << "t.numberOfAttackerPacketsAccepted=" << numberOfAttackerPacketsAccepted << std::endl;
			std::cout << "t.numberOfUnknownDataAccepted=" << numberOfUnknownDataAccepted << std::endl;
			std::cout << "t.numberOfPacketsSentByAttacker=" << numberOfPacketsSentByAttacker << std::endl;
			std::cout << "t.numberOfPacketsSentByRover=" << numberOfPacketsSentByRover << std::endl;
			std::cout << "t.numberOfPacketsSentByEarth=" << numberOfPacketsSentByEarth << std::endl;
			std::cout << "t.packetReceivedProgram_activeGeneCount=" << packetReceivedProgram->activeGeneCount << std::endl;
			std::cout << "t.dataReadyProgram_activeGeneCount=" << dataReadyProgram->activeGeneCount << std::endl;
			std::cout << "t.packetExpiredProgram_activeGeneCount=" << packetExpiredProgram->activeGeneCount << std::endl;
			if(attackerPacketReceivedProgram) {
				std::cout << "t.attackerPacketReceivedProgram_activeGeneCount=" << attackerPacketReceivedProgram->activeGeneCount << std::endl;
			}
			std::cout << "t.endTrial" << std::endl;
		}
	}

	void RunValidation() {
		isValidation = true;
		Run(false);
		isValidation = false;
	}

	void Run(bool printFitnessDetails=false) {
		packetReceivedExecutor.Setup(packetReceivedProgram);
		dataReadyExecutor.Setup(dataReadyProgram);
		packetExpiredExecutor.Setup(packetExpiredProgram);
		if(attackerPacketReceivedProgram) {
			attackerPacketReceivedExecutor.Setup(attackerPacketReceivedProgram);
		}

		CreateDataSchedule();
		RunSimulation();
		CalculateFitnessScores(printFitnessDetails);
	}

	static float NextTime(float currentValue, float candidateValue) {
		if(candidateValue < currentValue) {
			return candidateValue;
		}
		return currentValue;
	}

	void RunSimulation() {
		now = 0;

		while(now < INFINITE_CLOCK && now < stopTime) {
			float whenNext = INFINITE_CLOCK;
			whenNext = NextTime(whenNext,DoDataReady());
			whenNext = NextTime(whenNext,DoTransmit(roverHardware,toEarthInSpaceQueue,NULL));
			whenNext = NextTime(whenNext,DoPacketReceived(earthHardware,toEarthInSpaceQueue,toRoverInSpaceQueue,attackerPacketReceivedProgram? &attacker.fromSpace : NULL)); // Data received by earth is received by the ground and by the attacker, with the attacker acting faster than the ground station.
			if(attackerPacketReceivedProgram) {
				whenNext = NextTime(whenNext,DoAttacker());
			}
			whenNext = NextTime(whenNext,DoTransmit(earthHardware,toRoverInSpaceQueue, attackerPacketReceivedProgram? &attacker.fromEarth : NULL)); // Transmitting from earth to the rover should enqueue a copy at the attacker.
			if(attackerPacketReceivedProgram) {
				whenNext = NextTime(whenNext,DoAttacker());
			}
			whenNext = NextTime(whenNext,DoPacketExpired(earthHardware));
			whenNext = NextTime(whenNext,DoPacketReceived(roverHardware,toRoverInSpaceQueue,toEarthInSpaceQueue,NULL));
			whenNext = NextTime(whenNext,DoPacketExpired(roverHardware));
			now = whenNext;
			//std::cout << "now=" << now << std::endl;
		}
	}

	float DoAttacker() {
		float whenNext = INFINITE_CLOCK;

		// Check the list of packets from space
		while(!attacker.fromSpace.empty()) {
			Packet *p = attacker.fromSpace.front();
			attacker.fromSpace.pop_front();
			DoAttacker(p);
		}
		// Check the list of packets from earth
		while(!attacker.fromEarth.empty()) {
			Packet *p = attacker.fromEarth.front();
			attacker.fromEarth.pop_front();
			DoAttacker(p);
		}
		// Check our last sent packet
		if(attacker.lastSentPacket) {
			if(now >= attacker.nextStepTime) {
				Packet *p = attacker.lastSentPacket;
				attacker.lastSentPacket = NULL;
				DoAttacker(p);
			} else {
				whenNext = NextTime(whenNext,attacker.nextStepTime);
			}
		}

		return whenNext;
	}

	void DoAttacker(Packet *p) {
		if(!attackerPacketReceivedProgram) {
			packetBin.Recycle(p);
			p = NULL;
			return;
		}
		// Process the packet through attackerPacketReceivedProgram
		// Set inputs
		int inputIndex = 0;
		for(int i=0;i<ATTACKER_STATE_COUNT;i++) {
			attackerPacketReceivedExecutor.SetInput(inputIndex++,attacker.attackerState[i]);
		}
		attackerPacketReceivedExecutor.SetInput(inputIndex++,(float)p->dataType);
		attackerPacketReceivedExecutor.SetInput(inputIndex++,(float)(p->source == DS_Attacker? -1 : (p->source == DS_Earth? 0 : 1))); // -1 == Attacker, 0 == Earth, 1 == Rover
		attackerPacketReceivedExecutor.SetInput(inputIndex++,p->signatureType);
		float signatureHasBeenCracked = 0;
		if(p->source != DS_Attacker) {
			attacker.OnSignatureTypeSeen((int)p->signatureType,p->dataType,randomSource);
			signatureHasBeenCracked = (float)(attacker.IsSignatureTypeCracked((int)p->signatureType)? 1 : 0);
		}
		attackerPacketReceivedExecutor.SetInput(inputIndex++,signatureHasBeenCracked);
		attackerPacketReceivedExecutor.SetInput(inputIndex++,p->sequenceNumber);
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			attackerPacketReceivedExecutor.SetInput(inputIndex++,p->controlValues[i]);
		}

		// Run the CGP logic
		attackerPacketReceivedExecutor.Execute(attackerPacketReceivedProgram);

		// Collect outputs
		int outputIndex = 0;
		for(int i=0;i<ATTACKER_STATE_COUNT;i++) {
			attacker.attackerState[i] = attackerPacketReceivedExecutor.GetOutput(outputIndex++);
		}
		DataType dataTypeToSend = (DataType)(abs((int)attackerPacketReceivedExecutor.GetOutput(outputIndex++)) % 4);
		float dataSource = attackerPacketReceivedExecutor.GetOutput(outputIndex++);
		float signatureType = (float)((int)attackerPacketReceivedExecutor.GetOutput(outputIndex++));
		//signatureHasBeenCracked = attacker.IsSignatureTypeCracked((int)signatureType)? 1 : 0;
		float sequenceNumber = (float)((int)attackerPacketReceivedExecutor.GetOutput(outputIndex++));
		float sendPacket = attackerPacketReceivedExecutor.GetOutput(outputIndex++);
		DataSource destination = attackerPacketReceivedExecutor.GetOutput(outputIndex++) < 1? DS_Earth : DS_Rover;

		if(sendPacket < 1) { // If not sending a packet in response, discard the incoming packet and be done.
			// Done with the incoming packet
			packetBin.Recycle(p);
			p = NULL;
			return;
		}

		// Reuse the incoming packet
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			p->controlValues[i] = attackerPacketReceivedExecutor.GetOutput(outputIndex++);
		}
		p->source = DS_Attacker;
		p->sequenceNumber = sequenceNumber;
		p->signatureType = signatureType;
		p->signatureOk = attacker.IsSignatureTypeCracked((int)signatureType);
		if(dataSource <= 0 && p->dataType != DT_None) { // If the desire is to clone the incoming data, reuse its data type and ID
			// Keep existing p->dataType and ignore
			// Keep existing p->dataID
		} else if(dataTypeToSend != DT_None) { // The desire is to send fake data
			p->dataID = (float)(randomSource.NextInt(dataScheduleCount) + 1);
			p->dataType = dataTypeToSend;
		} else {
			p->dataID = 0;
			p->dataType = DT_None;
		}
		numberOfPacketsSentByAttacker++;

		if(destination == DS_Rover) {
			SpacePacket *sp = spacePacketBin.New();
			sp->packet = p;
			sp->whenArrives = now + SPACE_SECONDS;
			toRoverInSpaceQueue.push_back(sp);
			attackerPacketReceivedFitness += PENALTY_ATTACKER_SENT_TO_MOON;
		} else if(destination == DS_Earth) {
			SpacePacket *sp = spacePacketBin.New();
			sp->packet = p;
			sp->whenArrives = now + SPACE_SECONDS; // The attacker times their response to the earth to match the rover's delay so as not to raise suspicion
			toEarthInSpaceQueue.push_back(sp);
			attackerPacketReceivedFitness += PENALTY_ATTACKER_SENT_TO_EARTH;
		} else {
			// Done with the incoming packet
			packetBin.Recycle(p);
			p = NULL;
			return;
		}

		// If sending a new packet, clone the packet being sent and hold onto the clone waiting to be executed again.
		// This ensures the attacker gets to execute the next step in a strategy allowing the packet to be part of the attacker's state.
		if(!attacker.lastSentPacket) {
			attacker.lastSentPacket = packetBin.New();
		}
		*attacker.lastSentPacket = *p;
		attacker.nextStepTime = now + ATTACKER_IDLE_SECONDS;

		// Packet p is already owned by a SpacePacket at this point, so don't recycle it.
	}

	float DoDataReady() {
		while(nextDataIndex < dataScheduleCount && dataSchedule[nextDataIndex].creationTime <= now) {
			//std::cout << "data is ready to send" << std::endl;
			DataLump *d = &dataSchedule[nextDataIndex];
			d->hasBeenDequeued = true;
			Transceiver *transceiver = d->whichTransceiver == DS_Earth? &earthHardware : &roverHardware;
			if(transceiver->transmitQueue.size() < TRANSMIT_QUEUE_COUNT) {
				// Set inputs
				int inputIndex = 0;
				dataReadyExecutor.SetInput(inputIndex++,transceiver->transmitterState);
				for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
					dataReadyExecutor.SetInput(inputIndex++,transceiver->sharedState[i]);
				}
				dataReadyExecutor.SetInput(inputIndex++,(float)d->dataType);
				dataReadyExecutor.SetInput(inputIndex++,transceiver->nextSequenceNumber);
				dataReadyExecutor.SetInput(inputIndex++,transceiver->GetOldestSequenceInHistory());
				// Run the CGP logic
				dataReadyExecutor.Execute(dataReadyProgram);
				// Collect outputs
				Packet *p = packetBin.New();
				int outputIndex = 0;
				transceiver->transmitterState = dataReadyExecutor.GetOutput(outputIndex++);
				for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
					transceiver->sharedState[i] = dataReadyExecutor.GetOutput(outputIndex++);
				}
				p->signatureType = (float)((int)dataReadyExecutor.GetOutput(outputIndex++));
				for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
					p->controlValues[i] = dataReadyExecutor.GetOutput(outputIndex++);
				}
				// Complete the packet and enqueue it
				p->signatureOk = true;
				p->dataID = d->id;
				p->dataType = d->dataType;
				p->sequenceNumber = transceiver->nextSequenceNumber;
				transceiver->nextSequenceNumber += transceiver->sequenceIncrement;
				p->source = d->whichTransceiver;
				switch(p->source) {
					case DS_Earth:
						numberOfPacketsSentByEarth++;
						break;
					case DS_Rover:
						numberOfPacketsSentByRover++;
						break;
				}
				p->whenExpiresInHistory = now;
				transceiver->transmitQueue.push_back(p);
				// Queue a clone of data packets into the history along with a timeout
				// Only queue clones of DT_Command packets.  Video and telemetry packets are streamed with no room on the rover to resend them.
				if(p->dataType == DT_Command && transceiver->historyQueue.size() < HISTORY_QUEUE_COUNT) {
					Packet *clone = packetBin.New();
					*clone = *p;
					clone->whenExpiresInHistory = now + TRANSMIT_HISTORY_TIMEOUT_SECONDS;
					transceiver->historyQueue.push_back(clone);
				}
			}
			nextDataIndex++;
		}
		if(nextDataIndex < dataScheduleCount && dataSchedule[nextDataIndex].creationTime > now) {
			return dataSchedule[nextDataIndex].creationTime;
		}
		return INFINITE_CLOCK;
	}

	float DoTransmit(Transceiver &transceiver, list<SpacePacket *> &toSpacePackets, list<Packet *> *copyOfSentPackets) {
		if(transceiver.packetInTransmitter && now >= transceiver.whenTransmitFinishes) {
			// Put the packet into outer space
			//std::cout << "putting a packet into space" << std::endl;
			if(copyOfSentPackets && transceiver.packetInTransmitter->source != DS_Attacker) {
				Packet *cp = packetBin.New();
				*cp = *transceiver.packetInTransmitter;
				copyOfSentPackets->push_back(cp);
			}
			SpacePacket *sp = spacePacketBin.New();
			sp->packet = transceiver.packetInTransmitter;
			sp->whenArrives = now + SPACE_SECONDS;
			toSpacePackets.push_back(sp);
			transceiver.packetInTransmitter = NULL;
			transceiver.whenTransmitFinishes = 0;
		}
		if(!transceiver.packetInTransmitter && !transceiver.transmitQueue.empty()) {
			//std::cout << "starting transmit" << std::endl;
			transceiver.packetInTransmitter = transceiver.transmitQueue.front();
			transceiver.transmitQueue.pop_front();
			transceiver.whenTransmitFinishes = now + TRANSMIT_SECONDS;
		}
		if(transceiver.packetInTransmitter) {
			return transceiver.whenTransmitFinishes;
		}
		return INFINITE_CLOCK;
	}

	float DoPacketReceived(Transceiver &transceiver, list<SpacePacket *> &fromSpacePackets, list<SpacePacket *> &toSpacePackets, list<Packet *> *copyOfReceivedPackets) {
		while(true) {
			if(fromSpacePackets.empty()) {
				return INFINITE_CLOCK;
			}
			SpacePacket *sp = fromSpacePackets.front();
			if(now < sp->whenArrives) {
				return sp->whenArrives;
			}
			//std::cout << "receiving a packet from space" << std::endl;
			fromSpacePackets.pop_front();
			// Lose some packets in space
			bool shouldLosePacket = false;
			if(sp->packet->dataType == DT_Command) {
				switch(spaceState) {
					case 0: // waiting for the first command packet
						spaceState=1;
						break;
					case 1: // sent the first command packet
						spaceState=2;
						shouldLosePacket=true;
						break;
					case 2: // sent the second command packet and beyond
						// Nothing more to do
						break;
				}
			}
			const float lossFraction = (float)(isValidation? VALIDATION_SPACE_LOSS_FRACTION : SPACE_LOSS_FRACTION);
			const float echoFraction = (float)(isValidation? VALIDATION_SPACE_ECHO_FRACTION : SPACE_ECHO_FRACTION);
			const float replayFraction = (float)(isValidation? VALIDATION_SPACE_REPLAY_FRACTION : SPACE_REPLAY_FRACTION);

			if(!shouldLosePacket && randomSource.Next() <= lossFraction) {
				shouldLosePacket = true;
			}
			if(shouldLosePacket) {
				if(sp->packet->dataType != DT_None) {
					DataLump *d = FindDataLump(sp->packet->dataID);
					if(d) {
						d->wasDiscardedBySpace = true;
					}
				}
				packetBin.Recycle(sp->packet);
				sp->packet = NULL;
				spacePacketBin.Recycle(sp);
				numberOfPacketsLostInSpace++;
			} else {
				Packet *p = sp->packet;

				if(randomSource.Next() <= echoFraction) {
					Packet *p2 = packetBin.New();
					*p2 = *p;
					sp->packet = p2;
					sp->whenArrives = now + SPACE_SECONDS;
					toSpacePackets.push_back(sp);
				} else {
					sp->packet = NULL;
					spacePacketBin.Recycle(sp);
				}
				sp = NULL;

				list<Packet *> tempPackets;
				tempPackets.push_back(p);
				if(randomSource.Next() <= replayFraction) {
					const int howManyReplay = randomSource.NextInt(5) + 1;
					for(int i=0;i<howManyReplay;i++) {
						Packet *p2 = packetBin.New();
						*p2 = *p;
						tempPackets.push_back(p2);
					}
				}
				while(!tempPackets.empty()) {
					p = tempPackets.front();
					tempPackets.pop_front();

					if(copyOfReceivedPackets && p->source != DS_Attacker) {
						Packet *cp = packetBin.New();
						*cp = *p;
						copyOfReceivedPackets->push_back(cp);
					}

					// Reject packets that contain data inappropriate for the receiver.
					// Rovers don't accept video or telemetry.
					// Earth doesn't accept commands.
					if((transceiver.location == DS_Earth && p->dataType == DT_Command)
							|| (transceiver.location == DS_Rover && (p->dataType == DT_Video || p->dataType == DT_Telemetry))) {
						// Done with the packet p
						packetBin.Recycle(p);
						p = NULL;
						continue;
					}
					if(p->signatureOk) {
						const float corruptFraction = (float)(isValidation? VALIDATION_SPACE_CORRUPT_FRACTION : SPACE_CORRUPT_FRACTION);
						if(randomSource.Next() <= corruptFraction) {
							if(p->signatureType <= 0) { // The lack of a signature means the packet will be corrupted
								p->signatureOk = false;
							} else { // Signature can repair corruption, see if it is too corrupt
								const float unrepairableFraction = (float)(isValidation? VALIDATION_SPACE_UNREPAIRABLE_CORRUPT_FRACTION : SPACE_UNREPAIRABLE_CORRUPT_FRACTION);
								if(randomSource.Next() <= unrepairableFraction) {
									p->signatureOk = false;
								}
							}
						}
					}

					// Set inputs
					int inputIndex = 0;
					packetReceivedExecutor.SetInput(inputIndex++,transceiver.packetReceivedState);
					for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
						packetReceivedExecutor.SetInput(inputIndex++,transceiver.sharedState[i]);
					}
					packetReceivedExecutor.SetInput(inputIndex++,(float)p->dataType);
					for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
						packetReceivedExecutor.SetInput(inputIndex++,p->controlValues[i]);
					}
					packetReceivedExecutor.SetInput(inputIndex++,p->signatureOk);
					packetReceivedExecutor.SetInput(inputIndex++,p->sequenceNumber > (transceiver.sequenceIncrement + transceiver.highestSequenceSeen));
					packetReceivedExecutor.SetInput(inputIndex++,p->sequenceNumber == (transceiver.sequenceIncrement + transceiver.highestSequenceSeen));
					packetReceivedExecutor.SetInput(inputIndex++,p->sequenceNumber <= transceiver.highestSequenceSeen);
					packetReceivedExecutor.SetInput(inputIndex++,p->sequenceNumber);
					// Run the CGP logic
					packetReceivedExecutor.Execute(packetReceivedProgram);
					// Collect outputs
					int outputIndex = 0;
					transceiver.packetReceivedState = packetReceivedExecutor.GetOutput(outputIndex++);
					for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
						transceiver.sharedState[i] = packetReceivedExecutor.GetOutput(outputIndex++);
					}
					bool acceptData = packetReceivedExecutor.GetOutput(outputIndex++) >= 1;
					bool resetSequenceChecker = packetReceivedExecutor.GetOutput(outputIndex++) >= 1;
					float signatureType = (float)((int)packetReceivedExecutor.GetOutput(outputIndex++));
					int controlFlagIndex = outputIndex;
					outputIndex += CONTROL_VALUE_COUNT;
					bool sendControlPacket = packetReceivedExecutor.GetOutput(outputIndex++) >= 1;
					float clearSequenceNumber = packetReceivedExecutor.GetOutput(outputIndex++);
					float resendSequenceNumber = packetReceivedExecutor.GetOutput(outputIndex++);

					// Clear the named sequence number from the transmit history and queue
					if(clearSequenceNumber > 0) {
						Packet *qp = NULL;
						for(list<Packet *>::iterator i=transceiver.historyQueue.begin();i!=transceiver.historyQueue.end();i++) {
							if((int)(*i)->sequenceNumber == (int)clearSequenceNumber) {
								qp = *i;
								transceiver.historyQueue.erase(i);
								break;
							}
						}
						if(qp == NULL) {
							for(list<Packet *>::iterator i=transceiver.transmitQueue.begin();i!=transceiver.transmitQueue.end();i++) {
								if((int)(*i)->sequenceNumber == (int)clearSequenceNumber) {
									qp = *i;
									transceiver.transmitQueue.erase(i);
									break;
								}
							}
						}
						if(qp) {
							packetBin.Recycle(qp);
							qp = NULL;
						}
					}
					// Requeue a packet from the transmitter history.
					if(resendSequenceNumber > 0) {
						// Find the packet in the history, removing it.
						Packet *qp = NULL;
						for(list<Packet *>::iterator i=transceiver.historyQueue.begin();i!=transceiver.historyQueue.end();i++) {
							if((int)(*i)->sequenceNumber == (int)clearSequenceNumber) {
								qp = *i;
								transceiver.historyQueue.erase(i);
								break;
							}
						}
						// If there is room in the transmit queue, enqueue the packet otherwise discard it.
						if(qp) {
							if(transceiver.transmitQueue.size() < TRANSMIT_QUEUE_COUNT) {
								transceiver.transmitQueue.push_back(qp);
							} else {
								packetBin.Recycle(qp);
							}
						}
					}
					// Accept data and check it for validity, updating fitness accordingly
					if(acceptData && p->dataType != DT_None) {
						if(p->signatureType > 0 && !p->signatureOk) {
							packetReceivedFitness += PENALTY_ACCEPTED_BAD_SIGNATURE;
							numberOfBadSignaturesAccepted++;
							if(p->source == DS_Attacker) {
								attackerPacketReceivedFitness -= PENALTY_ACCEPTED_BAD_SIGNATURE;
							}
						}
						if(p->source == DS_Attacker) {
							packetReceivedFitness += PENALTY_ACCEPTED_ATTACKER_PACKET;
							numberOfAttackerPacketsAccepted++;
							attackerPacketReceivedFitness -= PENALTY_ACCEPTED_ATTACKER_PACKET;
						}
						// Find the data item and mark it received
						DataLump *d = FindDataLump(p->dataID);
						if(d) {
							d->acceptanceCounter++;
							d->wasDiscardedBySpace = false;
							if(p->dataType != d->dataType) {
								packetReceivedFitness += PENALTY_ACCEPTED_DATA_LUMP_WITH_WRONG_DATA_TYPE;
							} else if(p->dataType == DT_Command && p->source != DS_Attacker && d->acceptanceCounter == 1) {
								attackerPacketReceivedFitness += PENALTY_COMMAND_PACKET_GOT_THROUGH;
							}
							if(p->source == DS_Attacker) {
								if(d->acceptanceCounter > 1) {
									attackerPacketReceivedFitness -= PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE;
								}
								if(p->dataType != d->dataType) {
									attackerPacketReceivedFitness -= PENALTY_ACCEPTED_DATA_LUMP_WITH_WRONG_DATA_TYPE;
								}
							}
						} else {
							packetReceivedFitness += PENALTY_ACCEPTED_UNKNOWN_DATA_LUMP;
							numberOfUnknownDataAccepted++;
							if(p->source == DS_Attacker) {
								attackerPacketReceivedFitness -= PENALTY_ACCEPTED_UNKNOWN_DATA_LUMP;
							}
						}
					}
					// Done with the packet p
					packetBin.Recycle(p);
					p = NULL;
					// Send a control packet
					if(sendControlPacket) {
						if(transceiver.transmitQueue.size() < TRANSMIT_QUEUE_COUNT) {
							p = packetBin.New();
							for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
								p->controlValues[i] = packetReceivedExecutor.GetOutput(controlFlagIndex+i);
							}
							p->dataID = 0;
							p->dataType = DT_None;
							p->sequenceNumber = 0;
							p->signatureOk = true;
							p->signatureType = signatureType;
							p->source = &transceiver == &earthHardware? DS_Earth : DS_Rover;
							transceiver.transmitQueue.push_back(p);
						} else {
							// The packet was dropped and not sent.
						}
					}
				}
			}
		}
	}

	float DoPacketExpired(Transceiver &transceiver) {
		if(!transceiver.historyQueue.empty()) {
			Packet *p = transceiver.historyQueue.front();
			if(p->whenExpiresInHistory > now) {
				return p->whenExpiresInHistory;
			}
			//std::cout << "packet expired in history" << std::endl;
			transceiver.historyQueue.pop_front();
			if(transceiver.transmitQueue.size() < TRANSMIT_QUEUE_COUNT) {
				// Set inputs
				int inputIndex = 0;
				for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
					packetExpiredExecutor.SetInput(inputIndex++,transceiver.sharedState[i]);
				}
				for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
					packetExpiredExecutor.SetInput(inputIndex++,p->controlValues[i]);
				}
				// Run the CGP logic
				packetExpiredExecutor.Execute(packetExpiredProgram);
				// Collect outputs
				int outputIndex = 0;
				for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
					transceiver.sharedState[i] = packetExpiredExecutor.GetOutput(outputIndex++);
				}
				for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
					p->controlValues[i] = packetExpiredExecutor.GetOutput(outputIndex++);
				}
				bool resendPacket = packetExpiredExecutor.GetOutput(outputIndex++) >= 1;
				if(resendPacket) {
					transceiver.transmitQueue.push_back(p);
				} else {
					packetBin.Recycle(p);
				}
			} else {
				packetBin.Recycle(p);
			}
		}
		return INFINITE_CLOCK;
	}

	DataLump *FindDataLump(float id) const {
		for(int i=0;i<nextDataIndex;i++) {
			if(dataSchedule[i].id == id) {
				if(dataSchedule[i].hasBeenDequeued) { // Only return data items that have actually been created
					return &dataSchedule[i];
				} else {
					return NULL;
				}
			}
		}
		return NULL;
	}
};

class CGPGroup {
public:
	CGPPopulation *population;
	int ni, no, nr, nc, populationSize;

public:
	CGPGroup()
		: population(NULL), ni(0), no(0), nr(0), nc(0), populationSize(5) {
	}

	CGPGroup(int populationSizeToUse, int naToUse, int niToUse, int noToUse, int nrToUse, int ncToUse)
		: population(NULL) {
		Setup(populationSizeToUse, naToUse, niToUse, noToUse, nrToUse, ncToUse);
	}

	~CGPGroup() {
		if(population) {
			delete population;
			population = NULL;
		}
	}

	void Setup(int populationSizeToUse, int naToUse, int niToUse, int noToUse, int nrToUse, int ncToUse) {
		populationSize = populationSizeToUse;
		ni = niToUse;
		no = noToUse;
		nr = nrToUse;
		nc = ncToUse;
		population = new CGPPopulation(populationSize,2,ni,no,nr,nc);
	}

	void Randomize(RandomNumberSource &randomSource) const {
		population->Randomize(randomSource);
	}

	void ResetFitness() {
		for(int i=0;i<populationSize;i++) {
			population->population[i].isTested = false;
			population->population[i].fitness = 0;
		}
	}
};

class ProtocolExperiment1 {
public:
	unsigned long experimentID;
	RandomNumberSource randomSource;
	CGPMutator *mutator;
	int generationNumber;
	float *fitness;
	int populationSize;
	int attackerPopulationSize;
	CGPGroup *packetReceivedGroup;
	CGPGroup *dataReadyGroup;
	CGPGroup *packetExpiredGroup;

	CGPGroup *attackerPacketReceivedGroup;

	CGPIndividual *bestPacketReceivedProgram;
	CGPIndividual *bestDataReadyProgram;
	CGPIndividual *bestPacketExpiredProgram;

	CGPIndividual *bestAttackerPacketReceivedProgram;

	float bestFitness;
	int bestInGeneration;
	bool bestChanged;
	float validatedFitness;

	float bestAttackerFitness;
	int bestAttackerInGeneration;

	vector<string> inputNamesPacketReceived;
	vector<string> outputNamesPacketReceived;

	vector<string> inputNamesDataReady;
	vector<string> outputNamesDataReady;

	vector<string> inputNamesPacketExpired;
	vector<string> outputNamesPacketExpired;

	vector<string> inputNamesAttackerPacketReceived;
	vector<string> outputNamesAttackerPacketReceived;

public:
	ProtocolExperiment1(unsigned long experimentIDToUse)
		: experimentID(experimentIDToUse), mutator(NULL), populationSize(POPULATION_SIZE), attackerPopulationSize(ATTACKER_POPULATION_SIZE),
		packetReceivedGroup(NULL), dataReadyGroup(NULL), packetExpiredGroup(NULL),
		attackerPacketReceivedGroup(NULL),
		generationNumber(0),
		bestPacketReceivedProgram(NULL), bestDataReadyProgram(NULL), bestPacketExpiredProgram(NULL),
		bestAttackerPacketReceivedProgram(NULL),
		bestFitness(WORST_FITNESS), bestInGeneration(-1), bestChanged(false), validatedFitness(WORST_FITNESS),
		bestAttackerFitness(WORST_FITNESS), bestAttackerInGeneration(-1) {
		mutator = new CGPMutator(MUTATION_RATE,randomSource);

		// PacketReceived
		inputNamesPacketReceived.push_back("receiveState");
		for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sharedState%d",i);
			inputNamesPacketReceived.push_back(buffer);
		}
		inputNamesPacketReceived.push_back("dataType");
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"inputControl%d",i);
			inputNamesPacketReceived.push_back(buffer);
		}
		inputNamesPacketReceived.push_back("signatureOk");
		inputNamesPacketReceived.push_back("sequenceHigh");
		inputNamesPacketReceived.push_back("sequenceNext");
		inputNamesPacketReceived.push_back("sequenceLow");
		inputNamesPacketReceived.push_back("sequenceNumber");

		outputNamesPacketReceived.push_back("receiveState");
		for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sharedState%d",i);
			outputNamesPacketReceived.push_back(buffer);
		}
		outputNamesPacketReceived.push_back("accept");
		outputNamesPacketReceived.push_back("resetSequence");
		outputNamesPacketReceived.push_back("signatureType");
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sendControl%d",i);
			outputNamesPacketReceived.push_back(buffer);
		}
		outputNamesPacketReceived.push_back("send");
		outputNamesPacketReceived.push_back("clearSequence");
		outputNamesPacketReceived.push_back("resendSequence");

		int ni = (int)inputNamesPacketReceived.size(); //  1+INTERNAL_STATE_COUNT+1+CONTROL_VALUE_COUNT+5;
		int no = (int)outputNamesPacketReceived.size(); // 1+INTERNAL_STATE_COUNT+3+CONTROL_VALUE_COUNT+3;
		int nr = ni*2;
		int nc = nr;
		packetReceivedGroup = new CGPGroup(populationSize,2,ni,no,nr,nc);
		packetReceivedGroup->Randomize(randomSource);
		bestPacketReceivedProgram = new CGPIndividual(2,ni,no,nr,nc);

		// DataReady
		inputNamesDataReady.push_back("transmitState");
		for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sharedState%d",i);
			inputNamesDataReady.push_back(buffer);
		}
		inputNamesDataReady.push_back("dataType");
		inputNamesDataReady.push_back("nextSequence");
		inputNamesDataReady.push_back("oldestSequence");

		outputNamesDataReady.push_back("transmitState");
		for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sharedState%d",i);
			outputNamesDataReady.push_back(buffer);
		}
		outputNamesDataReady.push_back("signatureType");
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sendControl%d",i);
			outputNamesDataReady.push_back(buffer);
		}

		ni = (int)inputNamesDataReady.size(); // 1+INTERNAL_STATE_COUNT+3;
		no = (int)outputNamesDataReady.size(); // 1+INTERNAL_STATE_COUNT+1+CONTROL_VALUE_COUNT;
		nr = ni*2;
		nc = nr;
		dataReadyGroup = new CGPGroup(populationSize,2,ni,no,nr,nc);
		dataReadyGroup->Randomize(randomSource);
		bestDataReadyProgram = new CGPIndividual(2,ni,no,nr,nc);

		// PacketExpired
		for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sharedState%d",i);
			inputNamesPacketExpired.push_back(buffer);
		}
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sentControl%d",i);
			inputNamesPacketExpired.push_back(buffer);
		}

		for(int i=0;i<INTERNAL_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sharedState%d",i);
			outputNamesPacketExpired.push_back(buffer);
		}
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sendControl%d",i);
			outputNamesPacketExpired.push_back(buffer);
		}
		outputNamesPacketExpired.push_back("resend");

		ni = (int)inputNamesPacketExpired.size(); // INTERNAL_STATE_COUNT+CONTROL_VALUE_COUNT;
		no = (int)outputNamesPacketExpired.size(); // INTERNAL_STATE_COUNT+CONTROL_VALUE_COUNT+1;
		nr = ni*2;
		nc = nr;
		packetExpiredGroup = new CGPGroup(populationSize,2,ni,no,nr,nc);
		packetExpiredGroup->Randomize(randomSource);
		bestPacketExpiredProgram = new CGPIndividual(2,ni,no,nr,nc);

		// AttackerPacketReceived
		for(int i=0;i<ATTACKER_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"attackerState%d",i);
			inputNamesAttackerPacketReceived.push_back(buffer);
		}
		inputNamesAttackerPacketReceived.push_back("dataType");
		inputNamesAttackerPacketReceived.push_back("source"); // -1 == Attacker, 0 == Earth, 1 == Rover
		inputNamesAttackerPacketReceived.push_back("signatureType");
		inputNamesAttackerPacketReceived.push_back("signatureHasBeenCracked");
		inputNamesAttackerPacketReceived.push_back("sequenceNumber");
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"inputControl%d",i);
			inputNamesAttackerPacketReceived.push_back(buffer);
		}

		for(int i=0;i<ATTACKER_STATE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"attackerState%d",i);
			outputNamesAttackerPacketReceived.push_back(buffer);
		}
		outputNamesAttackerPacketReceived.push_back("dataTypeToSend");
		outputNamesAttackerPacketReceived.push_back("dataSource");
		outputNamesAttackerPacketReceived.push_back("signatureType");
		outputNamesAttackerPacketReceived.push_back("sequenceNumber");
		outputNamesAttackerPacketReceived.push_back("sendPacket");
		outputNamesAttackerPacketReceived.push_back("destination");
		for(int i=0;i<CONTROL_VALUE_COUNT;i++) {
			char buffer[50];
			sprintf(buffer,"sendControl%d",i);
			outputNamesAttackerPacketReceived.push_back(buffer);
		}

		ni = (int)inputNamesAttackerPacketReceived.size();
		no = (int)outputNamesAttackerPacketReceived.size();
		nr = max(ni*4,10);
		nc = nr;
		attackerPacketReceivedGroup = new CGPGroup(attackerPopulationSize,2,ni,no,nr,nc);
		attackerPacketReceivedGroup->Randomize(randomSource);
		bestAttackerPacketReceivedProgram = new CGPIndividual(2,ni,no,nr,nc);
	}

	~ProtocolExperiment1() {
		if(packetReceivedGroup) {
			delete packetReceivedGroup;
			packetReceivedGroup = NULL;
		}
		if(dataReadyGroup) {
			delete dataReadyGroup;
			dataReadyGroup = NULL;
		}
		if(packetExpiredGroup) {
			delete packetExpiredGroup;
			packetExpiredGroup = NULL;
		}
		if(attackerPacketReceivedGroup) {
			delete attackerPacketReceivedGroup;
			attackerPacketReceivedGroup = NULL;
		}
		if(mutator) {
			delete mutator;
			mutator = NULL;
		}
		if(bestPacketReceivedProgram) {
			delete bestPacketReceivedProgram;
			bestPacketReceivedProgram = NULL;
		}
		if(bestDataReadyProgram) {
			delete bestDataReadyProgram;
			bestDataReadyProgram = NULL;
		}
		if(bestPacketExpiredProgram) {
			delete bestPacketExpiredProgram;
			bestPacketExpiredProgram = NULL;
		}
		if(bestAttackerPacketReceivedProgram) {
			delete bestAttackerPacketReceivedProgram;
			bestAttackerPacketReceivedProgram = NULL;
		}
	}

	void PrintStats(bool includeDetailFiles=true) const {
		std::cout << "g.parentPacketReceived_fitness=" << packetReceivedGroup->population->population[0].fitness << std::endl;
		std::cout << "g.parentDataReady_fitness=" << dataReadyGroup->population->population[0].fitness << std::endl;
		std::cout << "g.parentPacketExpired_fitness=" << packetExpiredGroup->population->population[0].fitness << std::endl;
		std::cout << "g.bestInGeneration=" << bestInGeneration << std::endl;
		std::cout << "g.bestFitness=" << bestFitness << std::endl;
		std::cout << "g.validatedFitness=" << validatedFitness << std::endl;
#ifdef USE_ATTACKER
		std::cout << "g.parentAttackerPacketReceived_fitness=" << attackerPacketReceivedGroup->population->population[0].fitness << std::endl;
		std::cout << "g.bestAttackerInGeneration=" << bestAttackerInGeneration << std::endl;
		std::cout << "g.bestAttackerFitness=" << bestAttackerFitness << std::endl;
#endif

		if(includeDetailFiles) {
			CGPGraphViz visualizer;
			char fileName[250];
			sprintf(fileName,"CGP_%lu_%d_packetReceived.cgp",experimentID,generationNumber);
			packetReceivedGroup->population->population[0].Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_packetReceived.dot",experimentID,generationNumber);
			visualizer.Create(packetReceivedGroup->population->population[0],fileName,inputNamesPacketReceived,outputNamesPacketReceived);

			sprintf(fileName,"CGP_%lu_%d_dataReady.cgp",experimentID,generationNumber);
			dataReadyGroup->population->population[0].Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_dataReady.dot",experimentID,generationNumber);
			visualizer.Create(dataReadyGroup->population->population[0],fileName,inputNamesDataReady,outputNamesDataReady);

			sprintf(fileName,"CGP_%lu_%d_packetExpired.cgp",experimentID,generationNumber);
			packetExpiredGroup->population->population[0].Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_packetExpired.dot",experimentID,generationNumber);
			visualizer.Create(packetExpiredGroup->population->population[0],fileName,inputNamesPacketExpired,outputNamesPacketExpired);
#ifdef USE_ATTACKER
			sprintf(fileName,"CGP_%lu_%d_attackerPacketReceived.cgp",experimentID,generationNumber);
			attackerPacketReceivedGroup->population->population[0].Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_attackerPacketReceived.dot",experimentID,generationNumber);
			visualizer.Create(attackerPacketReceivedGroup->population->population[0],fileName,inputNamesAttackerPacketReceived,outputNamesAttackerPacketReceived);
#endif
		}
	}

	void Run(bool shouldPrintFitnessDetails) {
		bestChanged = false;
		int howManyToBeTested = populationSize * FITNESS_ITERATION_COUNT;
#ifdef USE_ATTACKER
		howManyToBeTested *= attackerPopulationSize;
#endif
		// Reset the fitness of each population member
		packetReceivedGroup->ResetFitness();
		dataReadyGroup->ResetFitness();
		packetExpiredGroup->ResetFitness();
		attackerPacketReceivedGroup->ResetFitness();

		if(howManyToBeTested <= 0) {
			return;
		}

		// Create simulated worlds, assigning a member of the population to each world and an optional attacker as well.
		World1 **worlds = new World1 *[howManyToBeTested];
		for(int wi=0;wi<howManyToBeTested;wi++) {
#ifdef USE_ATTACKER
			const int populationIndex = ((wi / FITNESS_ITERATION_COUNT) / attackerPopulationSize) % populationSize;
			const int attackerIndex = (wi / FITNESS_ITERATION_COUNT) % attackerPopulationSize;
#else
			const int populationIndex = (wi / FITNESS_ITERATION_COUNT) % populationSize;
#endif
			worlds[wi] = new World1;
			// Assign a member from the population to each world
			worlds[wi]->packetReceivedProgram = &packetReceivedGroup->population->population[populationIndex];
			worlds[wi]->dataReadyProgram = &dataReadyGroup->population->population[populationIndex];
			worlds[wi]->packetExpiredProgram = &packetExpiredGroup->population->population[populationIndex];
#ifdef USE_ATTACKER
			// Assign an attacker to each world.
			worlds[wi]->attackerPacketReceivedProgram = &attackerPacketReceivedGroup->population->population[attackerIndex];
#endif
		}

		// Execute each world
		#pragma omp parallel for
		for(int wi=0;wi<howManyToBeTested;wi++) {
			srand((generationNumber+1+wi+1+(int)time(NULL))+omp_get_thread_num());
			worlds[wi]->Run(wi == 0 && shouldPrintFitnessDetails);
		}

		// Apply fitness from each world to the population.
		// There may be more than one world for each population member, so don't do this summation in parallel.
		// Average the fitness scores so we can compare these scores to other models with different FITNESS_ITERATION_COUNT values.
		for(int wi=0;wi<howManyToBeTested;wi++) {
#ifdef USE_ATTACKER
			if(worlds[wi]->attackerPacketReceivedProgram) {
				worlds[wi]->attackerPacketReceivedProgram->fitness += worlds[wi]->attackerPacketReceivedFitness / FITNESS_ITERATION_COUNT / populationSize;
			}
			const int repeatCount = FITNESS_ITERATION_COUNT * attackerPopulationSize;
#else
			const int repeatCount = FITNESS_ITERATION_COUNT;
#endif
			worlds[wi]->packetReceivedProgram->fitness += worlds[wi]->packetReceivedFitness / repeatCount;
			worlds[wi]->dataReadyProgram->fitness += worlds[wi]->dataReadyFitness / repeatCount;
			worlds[wi]->packetExpiredProgram->fitness += worlds[wi]->packetExpiredFitness / repeatCount;
		}

		// Cleanup
		for(int wi=0;wi<howManyToBeTested;wi++) {
			delete worlds[wi];
			worlds[wi] = NULL;
		}
		delete worlds;

		// Find the best individual
		int bestIndex = -1;
		float bestFitnessSoFar = WORST_FITNESS;
		for(int i=0;i<populationSize;i++) {
			float fitness = packetReceivedGroup->population->population[i].fitness
					+ dataReadyGroup->population->population[i].fitness
					+ packetExpiredGroup->population->population[i].fitness;
			if(fitness < bestFitnessSoFar) {
				bestFitnessSoFar = fitness;
				bestIndex = i;
			}
		}
		if(bestFitnessSoFar < bestFitness) {
			bestChanged = true;
			bestFitness = bestFitnessSoFar;
			bestPacketReceivedProgram->CopyFrom(packetReceivedGroup->population->population[bestIndex]);
			bestDataReadyProgram->CopyFrom(dataReadyGroup->population->population[bestIndex]);
			bestPacketExpiredProgram->CopyFrom(packetExpiredGroup->population->population[bestIndex]);
			bestInGeneration = generationNumber;

			CGPExecutor packetReceivedEx;
			packetReceivedEx.Setup(bestPacketReceivedProgram);
			packetReceivedEx.Optimize(bestPacketReceivedProgram);

			CGPExecutor dataReadyEx;
			dataReadyEx.Setup(bestDataReadyProgram);
			dataReadyEx.Optimize(bestDataReadyProgram);

			CGPExecutor packetExpiredEx;
			packetExpiredEx.Setup(bestPacketExpiredProgram);
			packetExpiredEx.Optimize(bestPacketExpiredProgram);

			CGPGraphViz visualizer;
			char fileName[250];
			sprintf(fileName,"CGP_%lu_%d_packetReceived_best.cgp",experimentID,generationNumber);
			bestPacketReceivedProgram->Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_packetReceived_best.dot",experimentID,generationNumber);
			visualizer.Create(*bestPacketReceivedProgram,fileName,inputNamesPacketReceived,outputNamesPacketReceived);

			sprintf(fileName,"CGP_%lu_%d_dataReady_best.cgp",experimentID,generationNumber);
			bestDataReadyProgram->Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_dataReady_best.dot",experimentID,generationNumber);
			visualizer.Create(*bestDataReadyProgram,fileName,inputNamesDataReady,outputNamesDataReady);

			sprintf(fileName,"CGP_%lu_%d_packetExpired_best.cgp",experimentID,generationNumber);
			bestPacketExpiredProgram->Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_packetExpired_best.dot",experimentID,generationNumber);
			visualizer.Create(*bestPacketExpiredProgram,fileName,inputNamesPacketExpired,outputNamesPacketExpired);

			RunValidation();
		}

#ifdef USE_ATTACKER
		// Find the best attacker
		bestIndex = -1;
		bestFitnessSoFar = WORST_FITNESS;
		for(int i=0;i<populationSize;i++) {
			float fitness = attackerPacketReceivedGroup->population->population[i].fitness;
			if(fitness < bestFitnessSoFar) {
				bestFitnessSoFar = fitness;
				bestIndex = i;
			}
		}
		if(bestFitnessSoFar < bestAttackerFitness) {
			bestAttackerFitness = bestFitnessSoFar;
			bestAttackerPacketReceivedProgram->CopyFrom(attackerPacketReceivedGroup->population->population[bestIndex]);
			bestAttackerInGeneration = generationNumber;

			CGPExecutor attackerPacketReceivedEx;
			attackerPacketReceivedEx.Setup(bestAttackerPacketReceivedProgram);
			attackerPacketReceivedEx.Optimize(bestAttackerPacketReceivedProgram);

			CGPGraphViz visualizer;
			char fileName[250];
			sprintf(fileName,"CGP_%lu_%d_attackerPacketReceived_best.cgp",experimentID,generationNumber);
			bestAttackerPacketReceivedProgram->Save(fileName);
			sprintf(fileName,"CGP_%lu_%d_attackerPacketReceived_best.dot",experimentID,generationNumber);
			visualizer.Create(*bestAttackerPacketReceivedProgram,fileName,inputNamesAttackerPacketReceived,outputNamesAttackerPacketReceived);
		}
#endif
	}

	void PromoteBest() {
		// Advance the population using CGP rules.  This acts to promote the best chromosomes to form
		// a new "individual".  Since this individual is a mix of chromosomes, it will have never been
		// tested and must be itself evaluated in the next round.
		float mutationRate = bestInGeneration- generationNumber >= MUTATION_ACCELERATION_GENERATIONS? HIGH_MUTATION_RATE : MUTATION_RATE;
		mutator->ChangeRate(mutationRate);

		packetReceivedGroup->population->PromoteBest(*mutator);
		dataReadyGroup->population->PromoteBest(*mutator);
		packetExpiredGroup->population->PromoteBest(*mutator);
#ifdef USE_ATTACKER
		attackerPacketReceivedGroup->population->PromoteBest(*mutator);
#endif
	}

	void RunValidation() {
#ifdef _DEBUG
		const int howManyToBeTested = 1;
#else
		const int howManyToBeTested = 200;
#endif

		// Create simulated worlds, assigning a member of the population to each world.
		World1 **worlds = new World1 *[howManyToBeTested];
		for(int wi=0, i=0;i<howManyToBeTested;i++) {
			worlds[wi] = new World1;
			// Assign a member from the population to each world
			worlds[wi]->packetReceivedProgram = bestPacketReceivedProgram;
			worlds[wi]->dataReadyProgram = bestDataReadyProgram;
			worlds[wi]->packetExpiredProgram = bestPacketExpiredProgram;
			worlds[wi]->attackerPacketReceivedProgram = NULL;
			// Move to the next slot in the worlds to be simulated
			wi++;
		}

		// Execute each world
		#pragma omp parallel for
		for(int wi=0;wi<howManyToBeTested;wi++) {
			srand((generationNumber+1+wi+1+(int)time(NULL))+omp_get_thread_num());
			worlds[wi]->RunValidation();
		}

		// Apply fitness from each world to the population
		float totalFitness = 0;
		for(int wi=0;wi<howManyToBeTested;wi++) {
			totalFitness += worlds[wi]->packetReceivedFitness;
			totalFitness += worlds[wi]->dataReadyFitness;
			totalFitness += worlds[wi]->packetExpiredFitness;
			// No attacker used in this simulation
		}

		// Cleanup
		for(int wi=0;wi<howManyToBeTested;wi++) {
			delete worlds[wi];
			worlds[wi] = NULL;
		}
		delete worlds;

		// Done
		validatedFitness = totalFitness / howManyToBeTested;
	}
};

unsigned long GetID() {
	static unsigned long increment = 0;

	unsigned long t = ((((unsigned long)time(NULL)) ^ ((unsigned long)clock())) + increment) ^ rand();
	increment += rand();

	return t;
}

template <class T>
void PrintGeneration(const T &experiment) {
	std::cout << "*****************************" << std::endl;
	std::cout << "g.startGeneration" << std::endl;
	std::cout << "g.generation=" << experiment.generationNumber << std::endl;
}

template <class T>
void PrintStats(const T &experiment, clock_t start) {
	clock_t now = clock();
	experiment.PrintStats();
	double seconds = (now - start) / (double)CLOCKS_PER_SEC;
	std::cout << "g.generationsPerSecond=" << (1+experiment.generationNumber)/seconds << std::endl;
}

void PrintEndGeneration() {
	std::cout << "g.endGeneration" << std::endl;
}

void PrintFunctions() {
	std::cout << "f.startFunctions" << std::endl;
	int functionCount = CGP::GetFunctionCount();
	std::cout << "f.functionCount=" << functionCount << std::endl;
	for(int i=0;i<functionCount;i++) {
		std::cout << "f.f_" << i << "=\"" << CGP::GetFunctionName(i) << "\"" << std::endl;
	}
	std::cout << "f.endFunctions" << std::endl;
}

unsigned long PrintExperiment(const char *type) {
	unsigned long experimentID = GetID();

	std::cout << "e.startExperiment" << std::endl;
	std::cout << "e.id=" << experimentID << std::endl;
	std::cout << "e.functionSet=" << CGP::GetFunctionSet() << std::endl;
	std::cout << "e.clipping=" << CGP::GetClipping() << std::endl;
	std::cout << "e.experimentType=" << type << std::endl;
	std::cout << "e.codeVersion=" << __DATE__ << std::endl;
	std::cout << "e.modelVersion=" << MODEL_VERSION << std::endl;
	std::cout << "e.populationSize=" << POPULATION_SIZE << std::endl;
#ifdef USE_ATTACKER
	std::cout << "e.attackerPopulationSize=" << ATTACKER_POPULATION_SIZE << std::endl;
#endif
	std::cout << "e.mutationRate=" << MUTATION_RATE << std::endl;
	std::cout << "e.highMutationRate=" << HIGH_MUTATION_RATE << std::endl;
	std::cout << "e.mutationAccelerationGenerations=" << MUTATION_ACCELERATION_GENERATIONS << std::endl;
	std::cout << "e.FITNESS_ITERATION_COUNT=" << FITNESS_ITERATION_COUNT << std::endl;
	std::cout << "e.CONTROL_VALUE_COUNT=" << CONTROL_VALUE_COUNT << std::endl;
	std::cout << "e.TRANSMIT_QUEUE_COUNT=" << TRANSMIT_QUEUE_COUNT << std::endl;
	std::cout << "e.HISTORY_QUEUE_COUNT=" << HISTORY_QUEUE_COUNT << std::endl;
	std::cout << "e.INTERNAL_STATE_COUNT=" << INTERNAL_STATE_COUNT << std::endl;
	std::cout << "e.TRANSMIT_HISTORY_TIMEOUT_SECONDS=" << TRANSMIT_HISTORY_TIMEOUT_SECONDS << std::endl;
	std::cout << "e.TRANSMIT_SECONDS=" << TRANSMIT_SECONDS << std::endl;
	std::cout << "e.SPACE_SECONDS=" << SPACE_SECONDS << std::endl;
	std::cout << "e.SPACE_LOSS_FRACTION=" << SPACE_LOSS_FRACTION << std::endl;
	std::cout << "e.SPACE_REPLAY_FRACTION=" << SPACE_REPLAY_FRACTION << std::endl;
	std::cout << "e.SPACE_ECHO_FRACTION=" << SPACE_ECHO_FRACTION << std::endl;
	std::cout << "e.VALIDATION_DATA_SCHEDULE_COUNT=" << VALIDATION_DATA_SCHEDULE_COUNT << std::endl;
	std::cout << "e.VALIDATION_SPACE_LOSS_FRACTION=" << VALIDATION_SPACE_LOSS_FRACTION << std::endl;
	std::cout << "e.VALIDATION_SPACE_REPLAY_FRACTION=" << VALIDATION_SPACE_REPLAY_FRACTION << std::endl;
	std::cout << "e.VALIDATION_SPACE_ECHO_FRACTION=" << VALIDATION_SPACE_ECHO_FRACTION << std::endl;
	std::cout << "e.RUN_TIME_AFTER_LAST_DATA_SECONDS=" << RUN_TIME_AFTER_LAST_DATA_SECONDS << std::endl;
	std::cout << "e.DATA_SCHEDULE_COUNT=" << DATA_SCHEDULE_COUNT << std::endl;
	std::cout << "e.PENALTY_ACCEPTED_BAD_SIGNATURE=" << PENALTY_ACCEPTED_BAD_SIGNATURE << std::endl;
	std::cout << "e.PENALTY_ACCEPTED_ATTACKER_PACKET=" << PENALTY_ACCEPTED_ATTACKER_PACKET << std::endl;
	std::cout << "e.PENALTY_ACCEPTED_UNKNOWN_DATA_LUMP=" << PENALTY_ACCEPTED_UNKNOWN_DATA_LUMP << std::endl;
	std::cout << "e.PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE=" << PENALTY_ACCEPTED_DATA_LUMP_MORE_THAN_ONCE << std::endl;
	std::cout << "e.PENALTY_NEVER_ACCEPTED_DATA_LUMP=" << PENALTY_NEVER_ACCEPTED_DATA_LUMP << std::endl;
	std::cout << "e.PENALTY_ACCEPTED_DATA_LUMP_WITH_WRONG_DATA_TYPE=" << PENALTY_ACCEPTED_DATA_LUMP_WITH_WRONG_DATA_TYPE << std::endl;
	std::cout << "e.PENALTY_COMMAND_PACKET_GOT_THROUGH=" << PENALTY_COMMAND_PACKET_GOT_THROUGH << std::endl;
	std::cout << "e.PENALTY_ATTACKER_SENT_TO_EARTH=" << PENALTY_ATTACKER_SENT_TO_EARTH << std::endl;
	std::cout << "e.PENALTY_ATTACKER_SENT_TO_MOON=" << PENALTY_ATTACKER_SENT_TO_MOON << std::endl;

	int useChromosomes = 1;
#ifndef USE_CHROMOSOMES
	useChromosomes = 0;
#endif
	std::cout << "e.useChromosomes=" << useChromosomes << std::endl;

	int floatingPointConstants = 1;
#ifdef CGP_INTEGER_CONSTANTS
	floatingPointConstants = 0;
#endif
	std::cout << "e.floatingPointConstants=" << floatingPointConstants << std::endl;

	int optimizeForSize = OPTIMIZE_FOR_SIZE;
	std::cout << "e.optimizeForSize=" << optimizeForSize << std::endl;

	int useAttacker = 0;
#ifdef USE_ATTACKER
	useAttacker = 1;
#endif
	std::cout << "e.useAttacker=" << useAttacker << std::endl;
	std::cout << "e.ATTACKER_STATE_COUNT=" << ATTACKER_STATE_COUNT << std::endl;


	return experimentID;
}

void PrintEndExperiment() {
	std::cout << "e.endExperiment" << std::endl;
}

void RunExperiment1() {
	unsigned long experimentID = PrintExperiment("ProtocolExperiment1");

	ProtocolExperiment1 experiment(experimentID);

	std::cout << "e.packetReceivedProgram_ni=" << experiment.packetReceivedGroup->ni << std::endl;
	std::cout << "e.packetReceivedProgram_no=" << experiment.packetReceivedGroup->no << std::endl;
	std::cout << "e.packetReceivedProgram_nr=" << experiment.packetReceivedGroup->nr << std::endl;
	std::cout << "e.packetReceivedProgram_nc=" << experiment.packetReceivedGroup->nc << std::endl;

	std::cout << "e.dataReadyProgram_ni=" << experiment.dataReadyGroup->ni << std::endl;
	std::cout << "e.dataReadyProgram_no=" << experiment.dataReadyGroup->no << std::endl;
	std::cout << "e.dataReadyProgram_nr=" << experiment.dataReadyGroup->nr << std::endl;
	std::cout << "e.dataReadyProgram_nc=" << experiment.dataReadyGroup->nc << std::endl;

	std::cout << "e.packetExpiredProgram_ni=" << experiment.packetExpiredGroup->ni << std::endl;
	std::cout << "e.packetExpiredProgram_no=" << experiment.packetExpiredGroup->no << std::endl;
	std::cout << "e.packetExpiredProgram_nr=" << experiment.packetExpiredGroup->nr << std::endl;
	std::cout << "e.packetExpiredProgram_nc=" << experiment.packetExpiredGroup->nc << std::endl;

	PrintEndExperiment();
	PrintFunctions();

	clock_t start = clock();
#ifdef _DEBUG
	const int iMax = 1; // 200;
#else
	const int iMax = 1000; // 200000; // 1000000; // 250000; // 2000;
#endif
	bool didPrintRecently = false;

	for(int i=0;i<iMax;i++) {
		if(i > 0) {
			experiment.PromoteBest();
		}
		bool shouldPrintFitnessDetails = i == 0 || (i%1000) == 0 || i==(iMax-1) || i < 20;
		if(shouldPrintFitnessDetails) {
			PrintGeneration(experiment);
		}
		experiment.Run(shouldPrintFitnessDetails);
		if(experiment.bestChanged && !shouldPrintFitnessDetails) {
			shouldPrintFitnessDetails = true;
			PrintGeneration(experiment);
		}
		didPrintRecently = false;
		if(shouldPrintFitnessDetails) {
			didPrintRecently = true;
			PrintStats(experiment,start);
			PrintEndGeneration();
		}
		experiment.generationNumber++;
	}
	if(!didPrintRecently) {
		PrintGeneration(experiment);
		PrintStats(experiment,start);
		PrintEndGeneration();
	}
}

/*
bool HandleCommandLine(int argc, char *argv[]) {
	// -graphviz filename.cgp
	bool foundAnyMatches = false;
	for(int i=0;i<argc-1;i++) {
		if(!stricmp(argv[i],"-graphviz")) {
			foundAnyMatches = true;
			string outputDottyFileName(argv[i]);
			outputDottyFileName += ".dot";
			CGPGraphViz v;
			v.Convert(argv[i],outputDottyFileName.c_str());
		}
	}
	return foundAnyMatches;
}
*/

int main(int argc, char* argv[])
{
	//srand(22);
	srand((unsigned int)time(NULL));

	/*
	if(HandleCommandLine(argc,argv)) {
		return 0;
	}
	*/

	RunExperiment1();

	return 0;
}
